/*
 * Copyright (C) 2003-4 by Dusan Ferbas - DF Soft
 * <a href="www.dfsoft.cz">www.dfsoft.cz</a>
 *
 * Permission to use, copy, modify, and distribute this software
 * for any purpose and without fee is hereby granted, provided
 * that the above copyright notices appear in all copies and that
 * both the copyright notice and this permission notice appear in
 * supporting documentation and that the name of DF Soft may not be
 * used to endorse or promote products derived from this 
 * software without specific prior written permission.
 *
 * This software is provided "as is" without express or implied
 * warranty. We are not liable for anything caused by using this software.
 *
 * author <a href="mailto:rhlebak@dfsoft.cz">Rostislav Hlebak</a>
 */

/****************************************************************************
 *
 * File:         file_sys.c
 * Author(s):    Rostislav Hlebak
 * Project:      Charon SNMP Implementation
 * Description:  SPI DataFlash file system emulation
 *
 ****************************************************************************/

#include <sys/types.h>
#include <sys/heap.h>
#include <string.h>

#include "stddef.h"							//offsetof

#include "../eeparams/eeparams.h"

#define FILESYS_VARS

#include "file_sys.h"
#include "spi_flash.h"

static char FLASH_FILESYS_ID[4] = "RFS0";

//******************************************************************************
//      GLOBAL VARIABLES
//******************************************************************************


//******************************************************************************
//      FUNCTION DECLARATIONS
//******************************************************************************
static int flash_FileSearch(u_char * filename);
static void flash_FillStructure(Tflash_FILE * fd, int idx, u_char cmd, u_char * filename);
static void flash_SetToEOF(Tflash_FILE * fd);
static void flash_SetEOF(Tflash_FILE * fd, u_char value);



//******************************************************************************
//      MAJOR FUNCTIONS
//******************************************************************************
int flash_InitFileSys(void)
{
	if ((gFileTable = (TFileTableS *) NutHeapAllocClear(sizeof(TFileTableS))) == NULL)
		return -1;
	eeprom_get(FILE_SYS_EEPROM_OFFSET, (u_char *) gFileTable, sizeof(TFileTableS));
	gFileTable->file_state = WRITE_READY;
	if (memcmp(gFileTable->file_sysID, FLASH_FILESYS_ID, sizeof(FLASH_FILESYS_ID)))
	{
		flash_clear();
	}
	SPI_init_flash_port();
	return 0;
}

//******************************************************************************
int flash_EndFileSys(void)
{
	NutHeapFree(gFileTable);
	return 0;
}

//******************************************************************************
Tflash_FILE *flash_open(u_char * filename, u_char cmd)
{
	int     fileidx;
	Tflash_FILE *fd;
	if (*filename == 0)
		return NULL;
	if ((fd = (Tflash_FILE *) NutHeapAllocClear(sizeof(Tflash_FILE))) == NULL)
		return NULL;
	fileidx = flash_FileSearch(filename);
	if (cmd & F_OPENAS_READ)
	{
		if (fileidx >= 0)					//if file is found
		{
			flash_FillStructure(fd, fileidx, cmd, filename);
			return fd;
		}

	}
	else if (cmd & F_OPENAS_WRITE)
	{
		// check write access
		if ((gFileTable->file_state == WRITE_READY))
		{
			// there is a space in file table and file does not exist
			if ((gFileTable->file_count < FILE_COUNT) && (fileidx < 0))
			{
				flash_FillStructure(fd, gFileTable->file_count, cmd, filename);
				return fd;
			}
			else
				// if found file is last one
			if (fileidx == (gFileTable->file_count - 1))
			{
				flash_FillStructure(fd, fileidx, cmd, filename);
				flash_SetToEOF(fd);			// for adding next records skip to append point
				return fd;
			}
		}
	}
	NutHeapFree(fd);
	return NULL;
}

//******************************************************************************
int flash_close(Tflash_FILE * fd)
{
	if (fd->open_as & F_OPENAS_READ)
	{
		//     ;
	}
	else if (fd->open_as & F_OPENAS_WRITE)
	{
		unsigned int pages = fd->page_counter;
		unsigned int bytes = fd->buffer_counter;
		fd->file_recp->file_length = (long)pages *FLASH_PAGE_SIZE + bytes;
		if (bytes)
		{
			SPI_active_buffer_to_page(2, fd->file_recp->first_page + pages);
			pages++;
		}
		fd->file_recp->pages_occupied = pages;
		eeprom_put(FILE_SYS_EEPROM_OFFSET + offsetof(TFileTableS, file_rec[fd->file_index]),	// save record to EEPROM
				   (u_char *) & (gFileTable->file_rec[fd->file_index]),
				   sizeof(TFileRecS));

		gFileTable->pages_used = fd->file_recp->first_page + pages;	// store number of used pages
		eeprom_put(FILE_SYS_EEPROM_OFFSET + offsetof(TFileTableS, pages_used),	// also to eeprom
				   (u_char *) & (gFileTable->pages_used), sizeof(gFileTable->pages_used));

		if (fd->file_index == gFileTable->file_count - 1)	// in case file was added
		{
			eeprom_put(FILE_SYS_EEPROM_OFFSET + offsetof(TFileTableS, file_count),	//save state to eeprom
					   (u_char *) & (gFileTable->file_count),
					   sizeof(gFileTable->file_count));

		}
		gFileTable->file_state = WRITE_READY;	// enable write access

	}
	NutHeapFree(fd);						//release allocated memory
	return 0;
}

//******************************************************************************
int flash_read(Tflash_FILE * fd, u_char * buffer, u_long length)
{
	int     read_bytes;
	if (!(fd->open_as & F_OPENAS_READ))
		return ERR_READONLY;
	if (length > (read_bytes = SPI_read(1, buffer, length, &(fd->buffer_counter),
										&(fd->page_counter), fd->file_recp->first_page,
										fd->file_recp->file_length)))
		flash_SetEOF(fd, 1);
	return read_bytes;

}

//******************************************************************************
int flash_write(Tflash_FILE * fd, u_char * buffer, u_long length)
{
	if (!(fd->open_as & F_OPENAS_WRITE))
		return ERR_WRITEONLY;
	if (SPI_write(2, buffer, length, &(fd->buffer_counter), &(fd->page_counter),
				  fd->file_recp->first_page) < 0)
		return ERR_NOSPACE;
	return 0;
}

//******************************************************************************
long flash_length(Tflash_FILE * fd)			// return file length
{
	if (fd->open_as & F_OPENAS_READ)
		return fd->file_recp->file_length;
	if (fd->open_as & F_OPENAS_WRITE)
		return (fd->buffer_counter + fd->page_counter * FLASH_PAGE_SIZE);
	return ERR_FILESIZE;

}

//******************************************************************************
int flash_delete(Tflash_FILE * fd)			//delete last file
{
	if (fd->open_as & F_OPENAS_WRITE)		// only last file can be opened for write
	{
		gFileTable->pages_used = fd->file_recp->first_page;
		gFileTable->file_count--;
		// smaze zaznam o souboru
		memset(&(gFileTable->file_rec[fd->file_index]), 0, sizeof(TFileRecS));
		//ulozi ho do eeprom
		eeprom_put(FILE_SYS_EEPROM_OFFSET +
				   offsetof(TFileTableS, file_rec[fd->file_index]),
				   (u_char *) & (gFileTable->file_rec[fd->file_index]),
				   sizeof(TFileRecS));
		eeprom_put(FILE_SYS_EEPROM_OFFSET + offsetof(TFileTableS, pages_used),	// pages_used also to eeprom
				   (u_char *) & (gFileTable->pages_used), sizeof(gFileTable->pages_used));
		eeprom_put(FILE_SYS_EEPROM_OFFSET + offsetof(TFileTableS, file_count),	//save file_count state to eeprom
				   (u_char *) & (gFileTable->file_count), sizeof(gFileTable->file_count));

		gFileTable->file_state = WRITE_READY;
		NutHeapFree(fd);					//release allocated memory
		return 0;
	}
	return ERR_NOTLAST;

}

//******************************************************************************
int flash_clear(void)
{
	if (gFileTable->file_state == WRITE_READY)
	{
		// erase whole table in RAM
		memset(gFileTable, 0, sizeof(TFileTableS));
		memcpy(gFileTable->file_sysID, FLASH_FILESYS_ID, sizeof(FLASH_FILESYS_ID));
		// save into eeprom
		eeprom_put(FILE_SYS_EEPROM_OFFSET, (u_char *) gFileTable, sizeof(TFileTableS));
		return 0;
	}
	return -1;
}

//******************************************************************************
long flash_getfreemem(void)
{
	return (long)(FLASH_PAGE_COUNT - gFileTable->pages_used) * FLASH_PAGE_SIZE;
}

//******************************************************************************
int flash_EOF(Tflash_FILE * fd)
{
	if (fd->open_as & F_EOF)
		return 1;
	else
		return 0;
}


//******************************************************************************
//          AUXILLIARY FUNCTIONS
//******************************************************************************
int flash_FileSearch(u_char * filename)
{
	int     i;
	for (i = 0; i < gFileTable->file_count; i++)
	{
		if (!strcasecmp(gFileTable->file_rec[i].file_name, filename))
			return i;
	}
	return ERR_NOFILE;
}

//******************************************************************************
void flash_FillStructure(Tflash_FILE * fd, int idx, u_char cmd, u_char * filename)
{
	TFileRecS *my_file_recp = &(gFileTable->file_rec[idx]);
	if (cmd & F_OPENAS_WRITE)
	{
		if (idx == gFileTable->file_count)	// in case file was added
		{
			strncpy(my_file_recp->file_name, filename, FILE_NAME_LENGTH);	// fill in file name
			my_file_recp->first_page = gFileTable->pages_used;	// start of write area
			gFileTable->file_count++;		// actualize file count
		}
		gFileTable->file_state = WRITE_BUSY;	// disable write access for others
	}
	fd->file_recp = my_file_recp;
	fd->file_index = idx;
	fd->buffer_counter = 0;
	fd->page_counter = 0;
	fd->open_as = (0 | cmd);
}

//******************************************************************************
void flash_SetToEOF(Tflash_FILE * fd)
{
	TFileRecS *my_file_recp = fd->file_recp;
	fd->buffer_counter = my_file_recp->file_length % FLASH_PAGE_SIZE;	//set buffercounter behind last data
	fd->page_counter = my_file_recp->pages_occupied - 1;	//set to last page
	SPI_page_to_active_buffer(2, my_file_recp->first_page + fd->page_counter);
}

//******************************************************************************
void flash_SetEOF(Tflash_FILE * fd, u_char value)
{
	if (value)
		fd->open_as |= F_EOF;
	else
		fd->open_as &= ~F_EOF;
}

//******************************************************************************

//******************************************************************************
//          DEBUG FUNCTIONS
//******************************************************************************
void MemStat(void)
{
	printf_P(PSTR("\r\nFree heap mem:  %d bytes"), NutHeapAvailable());
	printf_P(PSTR("\r\nFree flash mem: %ld bytes"), flash_getfreemem());
}

//******************************************************************************
void display_filerec(TFileRecS * filerec, unsigned int idx)
{
	if (idx < gFileTable->file_count)
		printf_P(PSTR("\r\n*\t"));
	else
		printf_P(PSTR("\r\n\t"));

	printf_P(PSTR("%d.  %12s\t%4d\t%4d\t%6ld"), idx, filerec->file_name,
			 filerec->first_page, filerec->pages_occupied, filerec->file_length);
}

//******************************************************************************
void display_table(u_char cmd)
{

	printf_P(PSTR("\r\n pages_used: %d  file_count: %d  file_state:  %d\n\r"),
			 gFileTable->pages_used, gFileTable->file_count, gFileTable->file_state);

	unsigned int i, j;
	j = FILE_COUNT;
	if ((cmd != FULL_TABLE) && (gFileTable->file_count <= FILE_COUNT))
		j = gFileTable->file_count;

	for (i = 0; i < j; i++)
	{
		display_filerec(&gFileTable->file_rec[i], i);
	}
}
