/*
 * Copyright (C) 2003-4 by Dusan Ferbas - DF Soft
 * <a href="www.dfsoft.cz">www.dfsoft.cz</a>
 *
 * Permission to use, copy, modify, and distribute this software
 * for any purpose and without fee is hereby granted, provided
 * that the above copyright notices appear in all copies and that
 * both the copyright notice and this permission notice appear in
 * supporting documentation and that the name of DF Soft may not be
 * used to endorse or promote products derived from this 
 * software without specific prior written permission.
 *
 * This software is provided "as is" without express or implied
 * warranty. We are not liable for anything caused by using this software.
 *
 * author <a href="mailto:rhlebak@dfsoft.cz">Rostislav Hlebak</a>
 */

/****************************************************************************
 *
 * File:         spi_flash.c
 * Author(s):    Rostislav Hlebak, Dusan Ferbas
 * Project:      Charon SNMP Implementation
 * Description:  SPI DataFlash low level, based on Atmel AN:
 *				 http://www.atmel.com/dyn/resources/prod_documents/doc1456.pdf
 *
 ****************************************************************************/

#include "spi_flash.h"
#include <stdio.h>
#include <sys/timer.h>

//#define SPI_DEBUG

//******************************************************************************
//      GLOBAL VARIABLES
//******************************************************************************
unsigned int gLastReadPage;


//******************************************************************************
//      MAJOR FUNCTIONS
//******************************************************************************


void SPI_init_flash_port(void)
{
	gLastReadPage = 0xFFFF;

	DDRB = 0xB7;							// SPI Port initialisation
	// SCK, MISO, MOSI, CS, LED, WP , RDYBSY, RST
	// PB7, PB6, PB5, PB4, PB3, PB2 , PB1, PB0
	// O I O O O O I O
	// 1 0 1 1 1 1 0 1
	PORTB = 0xF9;							// all outputs high, inputs have pullups (LED is off)

	sbi(SPSR, SPI2X);
	//SPSR |= 1;  // 1 - doublespeed, 0 - singlespeed;
	outb(SPCR, BV(SPE) | BV(MSTR) | BV(CPOL) | BV(CPHA));
	//SPCR = 0x5C;
}


#ifndef SPI_DEBUG
//  #define SPI_wait(info)  while (!(SPSR & 0x80))
#define SPI_wait(info)	while (!(SPSR & BV(SPIF)))
#else
static void SPI_wait(char *info)
{
	int     cnt_tmp = 0;

	while (!(SPSR & 0x80))
		cnt_tmp++;							// wait for data transfer to be completed
	if (cnt_tmp)
		printf(".%s:%u", info, cnt_tmp);
}
#endif

int SPI_wait_for_ready(void)
{
	int     cnt_tmp = 0;
#if 0										// cannot use this for 8 pin chip
	while (!(PINB & DF_RDY_BUSY))
		cnt_tmp++;
#else
	u_char  status = 0;
	PORTB &= ~DF_CHIP_SELECT;				// enable DataFlash
	SPDR = STATUS_REGISTER;
	SPI_wait("11");							// wait for data transfer to be completed
	do
	{
		SPDR = 0;
		SPI_wait("11");

		status = SPDR;
		cnt_tmp++;
		//printf("\n\r %d : %X", cnt_tmp, status);
	}
	while (!(status & 0x80));
	PORTB |= DF_CHIP_SELECT;				// disable DataFlash
#endif
	//printf("\n\r wait:%u cyklu.  ",  cnt_tmp);
	return cnt_tmp;
}

//*************************************************************************************************
//                  Write
//*************************************************************************************************
void SPI_init_write_active_buff(unsigned char active_buffer, unsigned int buffer_counter)
{
	SPI_wait_for_ready();
	PORTB &= ~DF_CHIP_SELECT;				// enable DataFlash
	if (active_buffer == 1)
		SPDR = BUFFER_1_WRITE;
	else
		SPDR = BUFFER_2_WRITE;
	SPI_wait("z");							// wait for data transfer to be completed

	SPDR = 0x00;							// dont cares
	SPI_wait("z");							// wait for data transfer to be completed

	SPDR = (char)(buffer_counter >> 8);		// dont cares plus first two bits of buffer address
	SPI_wait("z");							// wait for data transfer to be completed

	SPDR = (char)buffer_counter;			// buffer address (max. 2^8 = 256 pages)
	SPI_wait("z");							// wait for data transfer to be completed
}

//*************************************************************************************************
int SPI_write(unsigned char active_buffer, unsigned char *my_buffer,
			  unsigned int buf_length, unsigned int *buffer_counter,
			  unsigned int *page_counter, unsigned int first_page)
{
	SPI_init_write_active_buff(active_buffer, *buffer_counter);	//enable DataFlash and init write to buffer

	unsigned int bytes_in_buffer = buf_length;
	while (buf_length--)
	{

		if (*buffer_counter >= FLASH_PAGE_SIZE)
		{
			PORTB |= DF_CHIP_SELECT;		// disable DataFlash
			*buffer_counter = 0;
			SPI_active_buffer_to_page(active_buffer, first_page + *page_counter);
			(*page_counter)++;
			if ((first_page + *page_counter) >= FLASH_PAGE_COUNT)
			{
				return -1;					// flash is full
			}
			SPI_init_write_active_buff(active_buffer, *buffer_counter);	//enable DataFlash and init write to buffer

		}
		SPDR = *my_buffer++;				// write data into SPI Data Register
		SPI_wait("z");
		(*buffer_counter)++;
	}
	PORTB |= DF_CHIP_SELECT;				// disable DataFlash
	return bytes_in_buffer;
}

//*************************************************************************************************
//*************************************************************************************************
void SPI_active_buffer_to_page(unsigned char active_buffer, unsigned int page_counter)
{
	SPI_wait_for_ready();
	PORTB &= ~DF_CHIP_SELECT;				// enable DataFlash

	if (active_buffer == 1)					// if buffer1 is the active buffer
	{
		SPDR = B1_TO_MM_PAGE_PROG_WITH_ERASE;	// write data from buffer1 to page
	}
	else
	{
		SPDR = B2_TO_MM_PAGE_PROG_WITH_ERASE;	// write data from buffer2 to page
	}

	SPI_wait("1111111111");					// wait for data transfer to be completed

	SPDR = (char)(page_counter >> PAGE_ADDRESS);
	SPI_wait("22222222222");				// wait for data transfer to be completed

	SPDR = (char)(page_counter << (8 - PAGE_ADDRESS));
	SPI_wait("33333333333");				// wait for data transfer to be completed

	SPDR = 0x00;							// dont cares
	SPI_wait("444444444444");				// wait for data transfer to be completed

	PORTB |= DF_CHIP_SELECT;				// disable DataFlash

	SPI_wait_for_ready();
}

//*************************************************************************************************



//*************************************************************************************************
//                  Read
//*************************************************************************************************
void SPI_init_read_active_buff(unsigned char active_buffer, unsigned int buffer_counter)
{
	SPI_wait_for_ready();

	PORTB &= ~DF_CHIP_SELECT;				// enable DataFlash
	if (active_buffer == 1)					// if buffer1 is the active buffer
	{
		SPDR = BUFFER_1_READ;				// read from buffer1
	}
	else									// else
	{
		SPDR = BUFFER_2_READ;				// read from buffer2
	}
	SPI_wait("z");							// wait for data transfer to be completed

	SPDR = 0x00;							// write dont care byte
	SPI_wait("z");							// wait for data transfer to be completed

	SPDR = (char)(buffer_counter >> 8);		// start at buffer address buffer_counter
	SPI_wait("z");							// wait for data transfer to be completed

	SPDR = (char)(buffer_counter);			// start at buffer address buffer_counter
	SPI_wait("z");							// wait for data transfer to be completed

	SPDR = 0x00;							// write dont care byte
	SPI_wait("z");							// wait for data transfer to be completed
}

//*************************************************************************************************

int SPI_read(unsigned char active_buffer, unsigned char *my_buffer,
			 unsigned int buf_length, unsigned int *buffer_counter,
			 unsigned int *page_counter, unsigned int first_page,
			 unsigned long file_length)
{
	long    position = (long)FLASH_PAGE_SIZE * (*page_counter) + *buffer_counter;
	long    rest = (file_length - position);
	int     bytes_in_buffer = buf_length;
	if (rest < (long)buf_length)
		bytes_in_buffer = buf_length = rest;


	if (gLastReadPage != *buffer_counter)
		SPI_page_to_active_buffer(active_buffer, first_page + *page_counter);

	SPI_init_read_active_buff(active_buffer, *buffer_counter);

	while (buf_length--)
	{

		if (*buffer_counter >= FLASH_PAGE_SIZE)
		{
			PORTB |= DF_CHIP_SELECT;		// disable DataFlash
			*buffer_counter = 0;
			(*page_counter)++;
			SPI_page_to_active_buffer(active_buffer, first_page + *page_counter);
			SPI_init_read_active_buff(active_buffer, *buffer_counter);	//enable DataFlash and init write to buffer

		}
		SPDR = 0x00;						// write dummy value to start register shift
		SPI_wait("z");
		*my_buffer++ = SPDR;				//store value
		(*buffer_counter)++;

	}
	PORTB |= DF_CHIP_SELECT;				// disable DataFlash
	return bytes_in_buffer;
}

//*************************************************************************************************
void SPI_page_to_active_buffer(unsigned char active_buffer, unsigned int page_counter)
{
	gLastReadPage = page_counter;			// set last readed page
	SPI_wait_for_ready();

	PORTB &= ~DF_CHIP_SELECT;				// enable DataFlash
	if (active_buffer == 1)					// if buffer1 is the active buffer
	{
		SPDR = MM_PAGE_TO_B1_XFER;			// transfer next page to buffer1
	}
	else									// else
	{
		SPDR = MM_PAGE_TO_B2_XFER;			// transfer next page to buffer2
	}
	SPI_wait("1");							// wait for data transfer to be completed

	SPDR = (char)(page_counter >> PAGE_ADDRESS);

	SPI_wait("2");							// wait for data transfer to be completed

	SPDR = (char)(page_counter << (8 - PAGE_ADDRESS));

	SPI_wait("3");							// wait for data transfer to be completed

	SPDR = 0x00;							// write dont care byte
	SPI_wait("4");							// wait for data transfer to be completed

	PORTB |= DF_CHIP_SELECT;				// disable DataFlash and start transaction
}

//*************************************************************************************************
void active_buffer_to_ram(unsigned char active_buffer, unsigned char *my_buffer,
						  unsigned int buf_length)
{
	SPI_wait_for_ready();

	PORTB &= ~DF_CHIP_SELECT;				// enable DataFlash
	if (active_buffer == 1)					// if buffer1 is the active buffer
	{
		SPDR = BUFFER_1_READ;				// read from buffer1
	}
	else									// else
	{
		SPDR = BUFFER_2_READ;				// read from buffer2
	}
	SPI_wait("z");							// wait for data transfer to be completed

	SPDR = 0x00;							// write dont care byte
	SPI_wait("z");							// wait for data transfer to be completed

	SPDR = 0x00;							// write dont care byte
	SPI_wait("z");							// wait for data transfer to be completed

	SPDR = 0x00;							// start at buffer address 0
	SPI_wait("z");							// wait for data transfer to be completed

	SPDR = 0x00;							// write dont care byte
	SPI_wait("z");							// wait for data transfer to be completed

	while ( /*buffer_counter < */ buf_length--)
	{
		SPDR = 0x00;						// write dummy value to start register shift
		SPI_wait("X");
		*(my_buffer++) = SPDR;				// play data from shift register
	}
	PORTB |= DF_CHIP_SELECT;				// disable DataFlash
}

//*************************************************************************************************
