/*
 * Copyright (C) 2003-4 by Dusan Ferbas - DF Soft
 * <a href="www.dfsoft.cz">www.dfsoft.cz</a>
 *
 * Permission to use, copy, modify, and distribute this software
 * for any purpose and without fee is hereby granted, provided
 * that the above copyright notices appear in all copies and that
 * both the copyright notice and this permission notice appear in
 * supporting documentation and that the name of DF Soft may not be
 * used to endorse or promote products derived from this
 * software without specific prior written permission.
 *
 * This software is provided "as is" without express or implied
 * warranty. We are not liable for anything caused by using this software.
 *
 * author <a href="mailto:dferbas@dfsoft.cz">Dusan Ferbas</a>
 */

/****************************************************************************
 *
 * File:         eeprom.c
 * Author(s):    Dusan Ferbas
 * Project:      WEB51 Charon SNMP Implementation
 * Description:  eeprom data upgrade/downgrade functionality
 *
 * Note: This code occupies approx. 1.5kB (according to data tables)
 *       and it is intended to be used for complex parameter structures.
 *       It can help you with upgrade of different code versions you have in field.
 *       Imagine e.g. 1 parameter added to 1.4, then another 2 params added to 1.6 etc.
 *       Then 1.7 comes and is able to upgrade both 1.4 and 1.6 while preserving
 *       parameters. Moreover you can then downgrade to 1.4, test behaviour
 *       and upgrade back to 1.7. All the time with user parameters. Unused
 *       parameters are thrown off, new ones get their defaults.
 ****************************************************************************/

//#define EE_DEBUG                //define to produce info about parameter loading
//#define EE_INFO                 //define to announce type of EEPROM operation


#include <string.h>							//memset(),..
#include <sys/types.h>
#include <sys/heap.h>
#include <sys/confnet.h>					//Nut OS net config (MAC offset)
#include <avr/wdt.h>

#define EEPROM_VARS
#include "eeparams.h"

#ifdef EE_DEBUG
#include "../dump.h"						//dump_char()
#endif

extern int SetupJumper;

u_char eeprom_get(u_short addr, u_char * bufp, u_short count)
{
#if 0										//DF 22.10.2003 - tested, works
	for (; count > 0; count--)
		*bufp++ = eeprom_read_byte((void *)addr++);
#else

	eeprom_read_block(bufp, (void *)addr, count);
#endif

	return (u_char) 1;
}

u_char eeprom_put(u_short addr, u_char * bufp, u_short count)
{
#ifdef EE_DEBUG
	printf_P(PSTR("\nWriting %u bytes:"), count);
	dump_char((char *)bufp, count);
#endif
#if 1
	/*
	 * //DF 25.10.2003 - tested, works
	 * for (; count > 0; count--)
	 * eeprom_write_byte((void *)addr++, *bufp++);
	 */

	//JS 30.10.2003
	//significantly speeds up data saving (if not all data are updated)
	for (; count > 0; count--)
	{
		if (!(count & 0x7F))
			wdt_reset();

		if (eeprom_read_byte((void *)addr) != *bufp)
			eeprom_write_byte((void *)addr, *bufp);
		addr++, bufp++;
	}

#else
	//new in AVR GCC libgcc.a 1.1 (we are currently using 1.0)
	eeprom_write_block(bufp, (void *)addr, count);
#endif

	return (u_char) 1;
}


//************************************
// EEPROM upgrade/downgrade routines
//************************************
ee_tbl_data ee_tbl_def_P PROGMEM = {
	{
	 EEPROM_TBL_PTRN,
	 offsetof(ee_tbl_data, e_elem),
	 EE_VAR_NUM,
	 {{{0x00, 0x0a, 0x59, 0x63, 0x08, 0x27}
	   }
	  }
	 /* default which will be overloaded, see below
	  * - loading MAC from non table version
	  * - preserving MAC  */
	 }
	,

	EE_VPTRN,

	{
	 {EE_VAR_BAUDRATE, CFG_ITEM_LEN(BaudRate), CFG_ITEM_OFFS(BaudRate)}
	 ,
	 {EE_VAR_DATA_LENGTH, CFG_ITEM_LEN(DataLength), CFG_ITEM_OFFS(DataLength)}
	 ,
	 {EE_VAR_STOP_BITS, CFG_ITEM_LEN(StopBits), CFG_ITEM_OFFS(StopBits)}
	 ,
	 {EE_VAR_PARITY, CFG_ITEM_LEN(Parity), CFG_ITEM_OFFS(Parity)}
	 ,
	 {EE_VAR_COMM_TIMEOUT, CFG_ITEM_LEN(CommTimeout), CFG_ITEM_OFFS(CommTimeout)}
	 ,

	 {EE_VAR_IPA, CFG_ITEM_LEN(aux_my_ip_addr), CFG_ITEM_OFFS(aux_my_ip_addr)}
	 ,
	 {EE_VAR_MASK, CFG_ITEM_LEN(aux_my_ip_mask), CFG_ITEM_OFFS(aux_my_ip_mask)}
	 ,
	 {EE_VAR_GW, CFG_ITEM_LEN(aux_gw_ip_addr), CFG_ITEM_OFFS(aux_gw_ip_addr)}
	 ,
	 {EE_VAR_HTTP_PORT, CFG_ITEM_LEN(http_port), CFG_ITEM_OFFS(http_port)}
	 ,
	 {EE_VAR_USER_SETTINGS, CFG_ITEM_LEN(UserSettings), CFG_ITEM_OFFS(UserSettings)}
	 }
};

/*--------------------------------------------------------------------------*/
/* returns data index of gBoardData in EEPROM */
static u_short eeprom_data_idx(register const ee_tbl * ee_tablep)
{
	return eeprom_tbl_tst(ee_tablep) ?
		EE_TABLEI + ee_tablep->t_offs + ee_tablep->t_count * sizeof(ee_tbl_elem) :
		EE_TABLEI;
}

/*--------------------------------------------------------------------------*/
/* looks for corresponding type in field of ee_tbl_elem structures
 * (will be compilated as inline) */
static u_char ee_tbl_idx(register const ee_tbl_elem * beg_elemp, u_short count,
						 register u_char type)
{
	register const ee_tbl_elem *end_elemp, *elemp;

	for (end_elemp = (elemp = beg_elemp) + count; elemp < end_elemp; elemp++)
	{
		if (elemp->v_type == type)
			return (elemp - beg_elemp);
	}

	return (u_char) - 1;					/* max number */
}

/*--------------------------------------------------------------------------*/
static u_char eeprom_upgrade(register const ee_tbl * old_tablep,
							 ee_tbl_data * ee_tbl_defp, u_char init_flg)
{
	u_char  tbl_upgrade[EE_VAR_NUM];
	/* these values has been read from previous table */
	u_char  ee_pass = 1;

	memset(tbl_upgrade, 0, sizeof(tbl_upgrade));
	/* assume nothing read from previous version */
	memset(gBoardDatap, 0, sizeof(BoardDataS));
	/* reset EEPROM contents to a known state (better for debugging) */

#ifdef EE_DEBUG
	puts_P(PSTR(", old_table:\n"));
	dump_char((char *)old_tablep, sizeof(ee_tbl));
#endif

	/* eeprom_idx indexes old data now */
	if (eeprom_idx == EE_TABLEI)			/* eeprom_tbl_tst(old_tablep)
											 * (EEPROM has no table structure) */
	{
		static const ETH_ADDR MAC_not_loaded = { {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF} };
		ETH_ADDR MAC_from_NutOS;

		/* place where Ethernut stores MAC */
		eeprom_get(CONFNET_EE_OFFSET + offsetof(CONFNET, cdn_mac),
				   (u_char *) & MAC_from_NutOS, sizeof(ETH_ADDR));

		/* check for fresh EEPROM */
		if (memcmp(&MAC_not_loaded, &MAC_from_NutOS, sizeof(ETH_ADDR)))
		{
#ifdef EE_DEBUG
			puts_P(PSTR("loading MAC from Nut net config place\n"));
#endif
			memcpy(&ee_tbl_defp->e_tbl.dta.enet_addr, &MAC_from_NutOS, sizeof(ETH_ADDR));
		}
		else
		{
			//TBD
			//          get_MAC((char*)&MAC_from_NutOS);
			if (memcmp(&MAC_not_loaded, &MAC_from_NutOS, sizeof(ETH_ADDR)))
			{
#ifdef EE_DEBUG
				puts_P(PSTR("loading MAC from Realtek EEPROM\n"));
#endif
				memcpy(&ee_tbl_defp->e_tbl.dta.enet_addr, &MAC_from_NutOS,
					   sizeof(ETH_ADDR));
			}

		}
	}
	else									//(eeprom_idx != EE_TABLEI)   /* (EEPROM has table structure) */
	{
		if (!init_flg)
		{
			register u_short ee_idx;
			ee_tbl_elem elem;
			register u_char elem_i;

#ifdef EE_DEBUG
			printf_P(PSTR("upgrading to new table (eeprom_idx:%d)\n"), eeprom_idx);
#endif

			/* browse old table (eeprom_idx set to old table data idx) */
			for (ee_idx = EE_TABLEI + old_tablep->t_offs; ee_idx < eeprom_idx;
				 ee_idx += sizeof(elem))
			{
				ee_pass &= eeprom_get(ee_idx, (u_char *) & elem, sizeof(elem));
#ifdef DBUG
				dump_char((char *)&elem, sizeof(elem));
#endif

				if (((elem_i =
					  ee_tbl_idx(&ee_tbl_defp->e_elem[0], EE_VAR_NUM,
								 elem.v_type)) != (u_char) - 1)
					&& (elem.v_len == ee_tbl_defp->e_elem[elem_i].v_len))
				{							/* variable from previous version found in current version
											 * and lengths equal */

					//elemp can be used to reduce code, but this is usefull only if EE_DEBUG
#ifdef EE_DEBUG
					printf_P(PSTR("upgrading variable %d:%u[%u] to :%u[%u]\n"),
							 (int)elem.v_type, elem.v_offs, (int)elem.v_len,
							 ee_tbl_defp->e_elem[elem_i].v_offs,
							 (int)ee_tbl_defp->e_elem[elem_i].v_len);
#endif
#ifdef EE_DEBUG
					if (tbl_upgrade[elem_i])
						printf_P(PSTR(" upgraded twice (%d)\n"), (int)elem.v_type);
#endif
#ifndef NO_CHANGE							/* test purposes (w/o this no upgrade will occure) */
					ee_pass &=
						eeprom_get(eeprom_idx + elem.v_offs,
								   (u_char *) gBoardDatap +
								   ee_tbl_defp->e_elem[elem_i].v_offs, elem.v_len);
					/* read directly into gBoardData structure */
#ifdef EE_DEBUG
					dump_char((char *)((u_char *) gBoardDatap +
									   ee_tbl_defp->e_elem[elem_i].v_offs), elem.v_len);
#endif


					tbl_upgrade[elem_i] = 1;
#else
					printf_P(PSTR("upgrading from %d[%u] to %.6lX"),
							 eeprom_idx + elem.v_offs, (int)elem.v_len,
							 (u_char *) gBoardDatap + ee_tbl_defp->e_elem[elem_i].v_offs);
#endif
				}
			}
		}
	}

	/* upgrade to new version */
	ee_pass &= eeprom_put(EE_TABLEI, (u_char *) ee_tbl_defp, sizeof(ee_tbl_def_P));
	eeprom_idx = eeprom_data_idx(&ee_tbl_defp->e_tbl);
#ifdef EE_DEBUG
	puts_P(PSTR("new table structure written to EEPROM\n"));
#endif

	{
		register int elem_i;

		/* browse new table and fill non upgraded variables */
		for (elem_i = 0; elem_i < EE_VAR_NUM; elem_i++)
		{
			register const ee_tbl_elem *elemp = &ee_tbl_defp->e_elem[elem_i];

#ifdef EE_DEBUG
			printf_P(PSTR("testing variable %d:%u[%u]"), (int)elemp->v_type,
					 elemp->v_offs, (int)elemp->v_len);

			if (tbl_upgrade[elem_i])
				puts_P(PSTR("\n"));
			else
#else
			if (!tbl_upgrade[elem_i])
#endif
			{
				static IP_ADDR ip_P PROGMEM = { {192, 168, 1, 99} };
				static IP_ADDR gw_P PROGMEM = { {192, 168, 1, 1} };
				static IP_ADDR mask_P PROGMEM = { {255, 255, 255, 0} };


#ifdef EE_DEBUG
				puts_P(PSTR(", setting\n"));
#endif
				switch (elemp->v_type)
				{
				case EE_VAR_IPA:
					memcpy_P(&gBoardDatap->aux_my_ip_addr, &ip_P, sizeof(IP_ADDR));
					break;

				case EE_VAR_MASK:
					memcpy_P(&gBoardDatap->aux_my_ip_mask, &mask_P, sizeof(IP_ADDR));
					break;

				case EE_VAR_GW:
					memcpy_P(&gBoardDatap->aux_gw_ip_addr, &gw_P, sizeof(IP_ADDR));
					break;
				case EE_VAR_HTTP_PORT:
					gBoardDatap->http_port = (u_short) 80;
				case EE_VAR_BAUDRATE:
					gBoardDatap->BaudRate = br9600;
					break;
				case EE_VAR_DATA_LENGTH:
					gBoardDatap->DataLength = db8;
					break;
				case EE_VAR_STOP_BITS:
					gBoardDatap->StopBits = sb1;
					break;
				case EE_VAR_PARITY:
					gBoardDatap->Parity = pNone;
					break;
				case EE_VAR_COMM_TIMEOUT:
					gBoardDatap->CommTimeout = 60;
					break;
				case EE_VAR_USER_SETTINGS:
					memset(&gBoardDatap->UserSettings, 0,
						   sizeof(gBoardDatap->UserSettings));
					break;
				default:
					break;
				}
			}
		}
	}

	/* initialize rest (everything that should be
	 * and was not transferred from previous version) */

	return ee_pass;
}

/* API for loading parameters to gBoardData */

u_char eeprom_load(u_char init_flg)
{
	/* obtain controller values */
	struct
	{
		ee_tbl  e_tbl;
		u_char  e_ptrn[EE_VPTRN_SIZE];
	}
	ee_data;
	u_char  ee_pass;

	ee_tbl_data *ee_tbl_defp;				/*  There are intensive comparisons to this table.
											 * This routine is called only during startup
											 * (plenty of heap, no other thread(s)
											 * was started yet)
											 * Also eeprom_put() is able to handle
											 * pointer from .data section (RAM area) only.
											 * Last - MAC address is modified if read from
											 * Nut NetConfig area.
											 * After that MAC address will be held
											 * at separate place because this large heap space
											 * will be freed.
											 */
	u_char  set_flg;

#ifdef EE_INFO
	//if (SetupJumper)      //DF 21.10.2003 - separate debug port on Charon II
	puts_P(PSTR("\nEEPROM loading, "));
#endif

	ee_pass = eeprom_get(EE_TABLEI, (u_char *) & ee_data, sizeof(ee_data));
	eeprom_idx = eeprom_data_idx(&ee_data.e_tbl);
	/* obtain old table data index */


	/* create place for holding parameters */
	gBoardDatap = (BoardDataS *) NutHeapAlloc(sizeof(BoardDataS));
	/* temporary create table that can be manipulated through standard rtns - e.g. memcmp
	 * (will be freed below) */
	ee_tbl_defp = (ee_tbl_data *) NutHeapAlloc(sizeof(ee_tbl_data /*ee_tbl_def_P */ ));
	memcpy_P(ee_tbl_defp, &ee_tbl_def_P, sizeof(ee_tbl_data /*ee_tbl_def_P */ ));

	/* preserve dta area (in case of init_flg below) */
	if (eeprom_idx != EE_TABLEI				/* eeprom_tbl_tst(&ee_data.e_tbl)
											 * (EEPROM has table structure) */
		)
	{
		memcpy(&ee_tbl_defp->e_tbl.dta, &ee_data.e_tbl.dta, sizeof(ee_dataS));
		/* replace table value with read value */

#ifdef EE_INFO
		//if (SetupJumper)
		puts_P(PSTR("preserving MAC, "));
#endif
	}


	if ((set_flg = (						//(init_flg = SetupJumper /*(InitJumper) or some switch combination*/)
					   init_flg || eeprom_idx == EE_TABLEI))	/* no table structure in EEPROM yet */
		|| memcmp(ee_data.e_ptrn, ee_tbl_defp->e_ptrn, sizeof(ee_data.e_ptrn)))
		/* older version or different application */
	{										/* data not valid yet */
#ifdef EE_INFO
		//if (SetupJumper)
		puts_P(set_flg ? PSTR("setting") : PSTR("upgrading"));
#endif

		ee_pass &= eeprom_upgrade(&ee_data.e_tbl, ee_tbl_defp, set_flg);
		/* either upgrade either set values */
		ee_pass &= eeprom_put(eeprom_idx, (u_char *) gBoardDatap, sizeof(BoardDataS));
		/* indicate validity */
	}
	else									/* get stored values */
	{
#ifdef EE_INFO
		//if (SetupJumper)
		puts_P(PSTR("getting"));
#endif

		ee_pass &= eeprom_get(eeprom_idx, (u_char *) gBoardDatap, sizeof(BoardDataS));
#ifdef EE_DEBUG
		dump_char((char *)gBoardDatap, sizeof(BoardDataS));
#endif
	}

#ifdef EE_INFO
	//if (SetupJumper)
	puts_P(PSTR("\n\n"));
#endif

	//preserve loaded MAC address
	memcpy(&g_ee_data, &ee_tbl_defp->e_tbl.dta, sizeof(ee_dataS));
	//release temporary data table (needed specially for AVR & Nut OS - 4kB .data)
	NutHeapFree(ee_tbl_defp);

	return ee_pass;
}

u_char eeprom_save(void)					// save changed data from gBoardData
{
	return eeprom_put(eeprom_idx, (u_char *) gBoardDatap, sizeof(BoardDataS));
}

int eeprom_verify(void)
{
	BoardDataS *VerBoardDatap;
	int     flagg = 0;
	u_char  ee_pass;

#ifdef EE_INFO
	puts_P(PSTR("\nEEPROM verifying, "));
#endif

	/* create place for holding parameters */
	VerBoardDatap = (BoardDataS *) NutHeapAlloc(sizeof(BoardDataS));
	/* temporary create table that can be manipulated through standard rtns - e.g. memcmp
	 * (will be freed below) */

	ee_pass = eeprom_get(eeprom_idx, (u_char *) VerBoardDatap, sizeof(BoardDataS));

	if (ee_pass)
		flagg = !memcmp(gBoardDatap, VerBoardDatap, sizeof(BoardDataS));
	NutHeapFree(VerBoardDatap);

	return (ee_pass & flagg);
}

int eeprom_clear(void)
{
	u_char *eepromp;
	u_char  ee_pass;
#define  ERASE_LEN (EE_TABLEI + sizeof(ee_tbl))

#ifdef EE_INFO
	puts_P(PSTR("\nEEPROM erasing, "));
#endif

	/* create place for holding parameters */
	eepromp = (u_char *) NutHeapAlloc(ERASE_LEN);
	memset(eepromp, 0xFF, ERASE_LEN);

	ee_pass = eeprom_put(0, eepromp, ERASE_LEN);

	NutHeapFree(eepromp);

	return ee_pass;
}
